@@ -311,7 +311,7 @@ sadump_generate_vmcoreinfo_from_vmlinux(size_t *vmcoreinfo_size)
 	}
 
 	strncpy(info->release, info->system_utsname.release,
-		strlen(info->system_utsname.release) + 1);
+		STRLEN_OSRELEASE);
 
 	write_vmcoreinfo_data();
 
@@ -1362,6 +1362,9 @@ finish:
  *    kernel. Retrieve vmcoreinfo from address of "elfcorehdr=" and
  *    get kaslr_offset and phys_base from vmcoreinfo.
  */
+#define PTI_USER_PGTABLE_BIT		(info->page_shift)
+#define PTI_USER_PGTABLE_MASK		(1 << PTI_USER_PGTABLE_BIT)
+#define CR3_PCID_MASK			0xFFFull
 int
 calc_kaslr_offset(void)
 {
@@ -1389,7 +1392,11 @@ calc_kaslr_offset(void)
 	}
 
 	idtr = ((uint64_t)smram.IdtUpper)<<32 | (uint64_t)smram.IdtLower;
-	cr3 = smram.Cr3;
+	if ((SYMBOL(pti_init) != NOT_FOUND_SYMBOL) ||
+	    (SYMBOL(kaiser_init) != NOT_FOUND_SYMBOL))
+		cr3 = smram.Cr3 & ~(CR3_PCID_MASK|PTI_USER_PGTABLE_MASK);
+	else
+		cr3 = smram.Cr3 & ~CR3_PCID_MASK;
 
 	/* Convert virtual address of IDT table to physical address */
 	if ((idtr_paddr = vtop4_x86_64_pagetable(idtr, cr3)) == NOT_PADDR)
@@ -2388,7 +2395,7 @@ sadump_kdump_backup_region_init(void)
 	elfcorehdr_p = 0;
 	for (i = 0; i < ARRAY_LENGTH(kimage.segment); ++i) {
 		char e_ident[EI_NIDENT];
-		unsigned mem;
+		unsigned long mem;
 
 		mem=ULONG(buf+i*SIZE(kexec_segment)+OFFSET(kexec_segment.mem));
 		if (!mem)
